import 'chore/ALDownloaderBatcherInputVO.dart';
import 'implementation/ALDownloaderBatcherIMP.dart';
import 'ALDownloader.dart';
import 'ALDownloaderHandlerInterface.dart';
import 'ALDownloaderStatus.dart';
import 'ALDownloaderTask.dart';
import 'ALDownloaderTypeDefine.dart';

/// ALDownloaderBatcher
///
/// The progress calculation method of [ALDownloaderHandlerInterface] is `progress = number of succeeded urls / number of all urls`.
abstract class ALDownloaderBatcher {
  /// Download
  ///
  /// **parameters**
  ///
  /// [urls] urls
  ///
  /// [downloaderHandlerInterface] downloader handler interface
  ///
  /// It is an one-off interface which will be destroyed when the download succeeded/failed.
  ///
  /// **return**
  ///
  /// [ALDownloaderHandlerInterfaceId] downloader handler interface id
  ///
  /// It is used for managing batch downloader handler interface.
  static ALDownloaderHandlerInterfaceId? download(List<String> urls,
          {ALDownloaderHandlerInterface? downloaderHandlerInterface}) =>
      ALDownloaderBatcherIMP.download(urls,
          downloaderHandlerInterface: downloaderHandlerInterface);

  /// Download
  ///
  /// **parameters**
  ///
  /// [vos] the input value object list
  ///
  /// [downloaderHandlerInterface] downloader handler interface
  ///
  /// It is an one-off interface which will be destroyed when the download succeeded/failed.
  ///
  /// **return**
  ///
  /// [ALDownloaderHandlerInterfaceId] downloader handler interface id
  ///
  /// It is used for managing downloader handler interface.
  static ALDownloaderHandlerInterfaceId? downloadUrlsWithVOs(
          List<ALDownloaderBatcherInputVO> vos,
          {ALDownloaderHandlerInterface? downloaderHandlerInterface}) =>
      ALDownloaderBatcherIMP.downloadUrlsWithVOs(vos,
          downloaderHandlerInterface: downloaderHandlerInterface);

  /// Add a downloader handler interface
  ///
  /// **parameters**
  ///
  /// [downloaderHandlerInterface] downloader handler interface
  ///
  /// It is an one-off interface which will be destroyed when the download succeeded/failed.
  ///
  /// [urls] urls
  ///
  /// **return**
  ///
  /// [ALDownloaderHandlerInterfaceId] downloader handler interface id
  ///
  /// It is used for managing downloader handler interface.
  static ALDownloaderHandlerInterfaceId addDownloaderHandlerInterface(
          ALDownloaderHandlerInterface downloaderHandlerInterface,
          List<String> urls) =>
      ALDownloaderBatcherIMP.addDownloaderHandlerInterface(
          downloaderHandlerInterface, urls);

  /// Remove downloader handler interface
  ///
  /// **parameters**
  ///
  /// [id] downloader handler interface id
  ///
  /// It is generated by [download], [downloadUrlsWithVOs], [addDownloaderHandlerInterface].
  static void removeDownloaderHandlerInterfaceForId(
          ALDownloaderHandlerInterfaceId id) =>
      ALDownloaderBatcherIMP.removeDownloaderHandlerInterfaceForId(id);

  /// Remove all downloader handler interfaces
  static void removeDownloaderHandlerInterfaceForAll() =>
      ALDownloaderBatcherIMP.removeDownloaderHandlerInterfaceForAll();

  /// Get download status
  ///
  /// Summarize the download status for a set of urls.
  ///
  /// **parameters**
  ///
  /// [urls] urls
  ///
  /// **return**
  ///
  /// [ALDownloaderStatus] status
  static Future<ALDownloaderStatus> getStatusForUrls(List<String> urls) =>
      ALDownloaderBatcherIMP.getStatusForUrls(urls);

  /// Get download progress
  ///
  /// number of succeeded urls / number of all urls
  ///
  /// **parameters**
  ///
  /// [urls] urls
  ///
  /// **return**
  ///
  /// [double] progress
  static Future<double> getProgressForUrls(List<String> urls) =>
      ALDownloaderBatcherIMP.getProgressForUrls(urls);

  /// Get tasks
  ///
  /// **parameters**
  ///
  /// [urls] urls
  ///
  /// [byQueueOrder]
  ///
  /// If [byQueueOrder] is true, return results in the order of tasks in the download queue,
  /// otherwise, in the order of [urls]'s order.
  ///
  /// **return**
  ///
  /// [List<ALDownloaderTask>] tasks
  static Future<List<ALDownloaderTask>> getTasksForUrls(List<String> urls,
          {bool byQueueOrder = false}) =>
      ALDownloaderBatcherIMP.getTasksForUrls(urls, byQueueOrder: byQueueOrder);

  /// Pause downloads
  ///
  /// This is a multiple of [ALDownloader.pause], see [ALDownloader.pause].
  ///
  /// **parameters**
  ///
  /// [urls] urls
  static void pause(List<String> urls) => ALDownloaderBatcherIMP.pause(urls);

  /// Cancel downloads
  ///
  /// This is a multiple of [ALDownloader.cancel], see [ALDownloader.cancel].
  ///
  /// **parameters**
  ///
  /// [urls] urls
  static void cancel(List<String> urls) => ALDownloaderBatcherIMP.cancel(urls);

  /// Remove downloads
  ///
  /// This is a multiple of [ALDownloader.remove], see [ALDownloader.remove].
  ///
  /// **parameters**
  ///
  /// [urls] urls
  static void remove(List<String> urls) => ALDownloaderBatcherIMP.remove(urls);

  /// Privatize constructor
  ALDownloaderBatcher._();
}
